;; -*- mode: common-lisp -*-
;;
;; Copyright (C) 2006 Franz Inc, Oakland, CA.  All rights reserved.
;;
;; This code is free software; you can redistribute it and/or
;; modify it under the terms of the version 2.1 of
;; the GNU Lesser General Public License as published by 
;; the Free Software Foundation, as clarified by the Franz
;; preamble to the LGPL found in
;; http://opensource.franz.com/preamble.html.
;;
;; This code is distributed in the hope that it will be useful,
;; but without any warranty; without even the implied warranty of
;; merchantability or fitness for a particular purpose.  See the GNU
;; Lesser General Public License for more details.
;;
;; Version 2.1 of the GNU Lesser General Public License can be
;; found at http://opensource.franz.com/license.html.
;; If it is not present, you can access it from
;; http://www.gnu.org/copyleft/lesser.txt (until superseded by a newer
;; version) or write to the Free Software Foundation, Inc., 59 Temple
;; Place, Suite 330, Boston, MA  02111-1307  USA
;;
;; $Id$

;; originally from ../ws/log.cl (refreshed on 1/27/05):

(defparameter *default-accesslog* (merge-pathnames (make-pathname :name "www_access")
						   *logroot*))
(defparameter *default-errorlog* (merge-pathnames (make-pathname :name "www_error")
						  *logroot*))

;; bind output streams to accesslog and errorlog to the specified hosts
;; (and default hosts, maybe) on server.
(defun bind-server-logs (servers &key hosts (accesslog *default-accesslog*)
					 (errorlog *default-errorlog*)
					 (default t))
  (when (atom servers) (setf servers (list servers)))
  (let (values-list
	(astr (net.aserve:ensure-stream-lock
	       (open accesslog :direction :output
		     :if-exists :append
		     :if-does-not-exist :create)))
	(estr (net.aserve:ensure-stream-lock
	       (open errorlog :direction :output
		     :if-exists :append
		     :if-does-not-exist :create))))
    (dolist (server servers)
      (let ((new-server (if (or (null server) (eq server :new))
			    (make-instance 'net.aserve:wserver)
			  server)))
	(dolist (host-name hosts)
	  (let ((host (gethash host-name (wserver-vhosts new-server))))
	    (setf (vhost-log-stream host) astr)
	    (setf (vhost-error-stream host) estr)))
	(when default
	  (let ((dh (wserver-default-vhost new-server)))
	    (setf (vhost-log-stream dh) astr)
	    (setf (vhost-error-stream dh) estr)))
	(push new-server values-list)))
    (values-list (reverse values-list))))

(defparameter *log-flush-thread* nil)

;; Start a thread to flush the log/error streams every 10 seconds.
(defun log-flush-thread ()
  (loop
    ;; Map across all vhosts on the default *server*
    (maphash (lambda (k v)
	       (declare (ignore k))
	       (let ((alog (vhost-log-stream v))
		     (elog (vhost-error-stream v)))
		 (if (and alog
			  (streamp alog)
			  (open-stream-p alog))
		     ;; Only operate on open streams.
		     (force-output alog))
		 (if (and elog
			  (streamp elog)
			  (open-stream-p elog))
		     ;; Only operate on open streams.
		     (force-output elog))))
	     (wserver-vhosts *server*))
    (sleep 10)))

;; Run the flusher if it's not running already.
(defun init-log-flush-thread ()
  (when (not *log-flush-thread*)
    (mp:process-run-function "log-flush-thread"
			     #'log-flush-thread)))

;; Publish-able function to flush and reopen log streams.
(defun flush-logs (req ent)
  ;; Flush them and reopen
  (close-and-reopen-logs)
  ;; Tell the client.
  (with-http-response (req ent)
    (with-http-body (req ent)
      (html (:html (:head (:title "Log flush complete"))
		   (:body "done."))))))

;; Flush log streams and reopen - needed for log rotation.
(defun close-and-reopen-logs ()
  (let ((streams nil)
	(old-streams nil)
	(vhosts nil))
    (flet (;; this function does the stream work.
	   (vhost-log-refresh (v)
	     (labels (;; reopen stream, use lock if available.
		      (reopen-lock (stream)
			(let ((lock
			       (and (streamp stream)
				    (getf
				     (excl::stream-property-list stream)
				     :lock))))
			  (if* lock
			     then ;; lock found, using it to reopen.
				  (mp:with-process-lock (lock)
				    (reopen stream))
			     else ;; no lock, reopening without one.
				  (reopen stream))))
		      ;; reopen stream if needed, called by reopen-lock.
		      (reopen (stream)
			(if* (cdr (assoc stream streams :test #'eq))
			   thenret ;;  Careful not to do the same stream
				   ;;; more than once.  This one's been
				   ;;; done already.
			 elseif (typep stream 'excl::file-simple-stream)
			   then (let ((path (pathname stream)))
				  (if* path
				     then (push stream old-streams)
					  (let ((new-stream 
						 (ensure-stream-lock
						  (open path
							:direction :output
							:if-exists :append
							:if-does-not-exist
							:create))))
					    ;; Add to list of done streams.
					    (push (cons stream new-stream)
						  streams)
					    new-stream)
				     else stream))
			   else stream)))
	       (if* (not (member v vhosts :test #'eq))
		  then ;; Only access each vhost once.
		       (setf (vhost-log-stream v)
			 (reopen-lock (vhost-log-stream v)))
		       (setf (vhost-error-stream v)
			 (reopen-lock (vhost-error-stream v)))
		       (push v vhosts)))))
      ;; refresh logs for all vhosts on *server*.
      (maphash (lambda (k v)
		 (declare (ignore k))
		 (vhost-log-refresh v))
	       (wserver-vhosts *server*))
      ;; refresh logs for all vhosts on *reload-server*.
      (maphash (lambda (k v)
		 (declare (ignore k))
		 (vhost-log-refresh v))
	       (wserver-vhosts *reload-server*))
      ;; have to do the default-vhosts separately, they're not in the
      ;; wserver-vhosts list.
      (vhost-log-refresh (wserver-default-vhost *server*))
      (vhost-log-refresh (wserver-default-vhost *reload-server*))
      ;; clean-up old streams.
      (dolist (elt old-streams)
	(let ((lock
	       (and (streamp elt)
		    (getf (excl::stream-property-list elt) :lock))))
	  (if* lock
	     then ;; lock found, close with it.
		  (mp:with-process-lock (lock)
		    (close elt))
	     else ;; no lock, close without.
		  (close elt)))))))

