;; -*- mode: common-lisp -*-
;;
;; Copyright (C) 2008 Franz Inc, Oakland, CA.  All rights reserved.
;;
;; This code is free software; you can redistribute it and/or
;; modify it under the terms of the version 2.1 of
;; the GNU Lesser General Public License as published by 
;; the Free Software Foundation, as clarified by the Franz
;; preamble to the LGPL found in
;; http://opensource.franz.com/preamble.html.
;;
;; This code is distributed in the hope that it will be useful,
;; but without any warranty; without even the implied warranty of
;; merchantability or fitness for a particular purpose.  See the GNU
;; Lesser General Public License for more details.
;;
;; Version 2.1 of the GNU Lesser General Public License can be
;; found at http://opensource.franz.com/license.html.
;; If it is not present, you can access it from
;; http://www.gnu.org/copyleft/lesser.txt (until superseded by a newer
;; version) or write to the Free Software Foundation, Inc., 59 Temple
;; Place, Suite 330, Boston, MA  02111-1307  USA
;;
;; $Id$

(in-package :user)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (require :ntservice))

(defvar *usage*
    "~
Usage: sample-server --register [--interact-with-desktop] [--port port] [-o logfile]
       sample-server --unregister
       sample-server --start-service
       sample-server --stop-service
       sample-server [--port port] [-o logfile] [--interact-with-desktop] [-d]
")

(defparameter *servicename* "aclwww")
(defparameter *displayname* "Allegro WWW")
(defparameter *description* "Common Lisp Web Service (powered by AllegroServe).")
(defparameter *commandline* "~a --port ~a~a")

(defun main ()
  (handler-case
      (sys:with-command-line-arguments 
	  (("port" :long port :required-companion)
	   ("register" :long register)
	   ("unregister" :long unregister)
	   ("start-service" :long start-service)
	   ("stop-service" :long stop-service)
	   ("interact-with-desktop" :long interact-p)
	   ("o" :short logfile :required-companion)
	   ("d" :short debug))
	  (rest :usage *usage*)
	(declare (ignorable rest))

	(unless logfile (setq logfile (merge-pathnames "www.out" *logroot*)))
	
	(when (or (and register unregister)
		  (and unregister start-service))
	  (error-die *usage*))

	(when port
	  (setq *port* (ignore-errors (parse-integer port :junk-allowed nil))))

	(when debug (break "debug away!"))

	(when (not (or interact-p register unregister start-service))
	  ;; app startup, non-interactive service
	  (setq *terminal-io* (open logfile :direction :output :if-exists :append
				    :if-does-not-exist :create)))
	  
	(when unregister
	  (remove-service)
	  (exit 0 :quiet t))
		
	(when start-service
	  (ntservice:start-service *servicename*)
	  (exit 0 :quiet t))
	
	(when stop-service
	  (ntservice:stop-service *servicename*)
	  (exit 0 :quiet t))
	
	(unless *port*
	  (error-die *usage*))
	
	(when register
	  (add-service *port* interact-p)
	  (exit 0 :quiet t))	
	
	(ntservice:execute-service *servicename* #'server-main
			 :stop 'stop-server))
    (error (c) (break "An error occurred: ~a." c))))

;; called by execute-service.
(defun server-main ()
  (init-server *port*))

(defun init-server (port)
  (format t "Starting www service at port ~a...~%" port)
  (multiple-value-setq (*server* *reload-server*) (bind-server-logs '(:new :new)))
  (start :port port :listeners 5 :server *server* :proxy t)
  (start :port (+ port 1000) :listeners 2 :server *reload-server*)
  
  (when (eql 80 port)
    (setf (wserver-filters *server*) '(filter-requests)))
  
  (mp:process-run-function "telnet server" 'start-telnet-server (+ port 1))
  
  (multiprocessing:process-allow-schedule)
  (reload)

  (loop (sleep most-positive-fixnum))
  ;; For maximum debugging information at the web server level, do this:
  #+ignore (net.aserve::debug-on :all))

(defun add-service (port interact-p)
  (format t "Installing service...~%")
  (multiple-value-bind (success errcode)
      (ntservice:create-service 
       *servicename*
       *displayname*
       (format nil *commandline* (sys:command-line-argument 0) port
	       (if interact-p " --interact-with-desktop" ""))
       :description *description*
       :interact-with-desktop interact-p)
    (if success
	(format t "Installation successful.~%")
      (error "ntservice:create-service error: ~A"
	     (ntservice:winstrerror errcode)))))

(defun remove-service ()
  (format t "Removing service...~%")
  (multiple-value-bind (success errcode errfunc)
      (ntservice:delete-service *servicename*)
    (if success
	(format t "Removal successful.~%")
      (error "ntservice:delete-service error in ~A: ~A"
	     errfunc (ntservice:winstrerror errcode)))))

(defun stop-server ()
  (format t "Shutting down server...~%")
  (shutdown :server *server*)
  (shutdown :server *reload-server*)
  (mp:process-kill (mp:process-name-to-process "telnet server")))
